@@ -3,14 +3,23 @@ import functools
 import json
 import logging
 import random
+import socket
 import urllib
 from urlparse import urlparse
 
-from radosgw_agent import request as aws_request
 from boto.exception import BotoServerError
 from boto.s3.connection import S3Connection
 
+from radosgw_agent import request as aws_request
+from radosgw_agent import config
+from radosgw_agent import exceptions as exc
+from radosgw_agent.util import get_dev_logger
+from radosgw_agent.constants import DEFAULT_TIME
+from radosgw_agent.exceptions import NetworkError
+
 log = logging.getLogger(__name__)
+dev_log = get_dev_logger(__name__)
+
 
 class Endpoint(object):
     def __init__(self, host, port, secure,
@@ -32,6 +41,7 @@ class Endpoint(object):
         # if self and other are mixed http/https with default ports,
         # i.e. http://example.com and https://example.com, consider
         # them the same
+
         def diff_only_default_ports(a, b):
             return a.secure and a.port == 443 and not b.secure and b.port == 80
         return (diff_only_default_ports(self, other) or
@@ -49,55 +59,36 @@ class Endpoint(object):
                                                  host=self.host,
                                                  port=self.port)
 
-class ClientException(Exception):
-    pass
-class InvalidProtocol(ClientException):
-    pass
-class InvalidHost(ClientException):
-    pass
-class InvalidZone(ClientException):
-    pass
-class ZoneNotFound(ClientException):
-    pass
-class BucketEmpty(ClientException):
-    pass
 
 def parse_endpoint(endpoint):
     url = urlparse(endpoint)
     if url.scheme not in ['http', 'https']:
-        raise InvalidProtocol('invalid protocol %r' % url.scheme)
+        raise exc.InvalidProtocol('invalid protocol %r' % url.scheme)
     if not url.hostname:
-        raise InvalidHost('no hostname in %r' % endpoint)
+        raise exc.InvalidHost('no hostname in %r' % endpoint)
     return Endpoint(url.hostname, url.port, url.scheme == 'https')
 
-class HttpError(ClientException):
-    def __init__(self, code, body):
-        self.code = code
-        self.str_code = str(code)
-        self.body = body
-        self.message = 'Http error code %s content %s' % (code, body)
-    def __str__(self):
-        return self.message
-class NotFound(HttpError):
-    pass
 code_to_exc = {
-    404: NotFound,
+    404: exc.NotFound,
     }
 
+
 def boto_call(func):
     @functools.wraps(func)
     def translate_exception(*args, **kwargs):
         try:
             return func(*args, **kwargs)
         except boto.exception.S3ResponseError as e:
-            raise code_to_exc.get(e.status, HttpError)(e.status, e.body)
+            raise code_to_exc.get(e.status, exc.HttpError)(e.status, e.body)
     return translate_exception
 
 
 def check_result_status(result):
     if result.status / 100 != 2:
         raise code_to_exc.get(result.status,
-                              HttpError)(result.status, result.reason)
+                              exc.HttpError)(result.status, result.reason)
+
+
 def url_safe(component):
     if isinstance(component, basestring):
         string = component.encode('utf8')
@@ -105,6 +96,7 @@ def url_safe(component):
         string = str(component)
     return urllib.quote(string)
 
+
 def request(connection, type_, resource, params=None, headers=None,
             data=None, expect_json=True, special_first_param=None, _retries=3):
     if headers is None:
@@ -134,6 +126,7 @@ def request(connection, type_, resource, params=None, headers=None,
 
     boto.log.debug('url = %r\nparams=%r\nheaders=%r\ndata=%r',
                    url, params, request.headers, data)
+
     try:
         result = aws_request.make_request(
             connection.s3_connection,
@@ -144,6 +137,10 @@ def request(connection, type_, resource, params=None, headers=None,
             data=request_data,
             params=safe_params,
             _retries=_retries)
+    except socket.error as error:
+        msg = 'unable to connect to %s %s' % (request.host, error)
+        raise NetworkError(msg)
+
     except BotoServerError as error:
         check_result_status(error)
 
@@ -184,7 +181,7 @@ def get_op_state(connection, client_id, op_id, bucket, obj):
     return request(connection, 'get', 'admin/opstate',
                    params={
                        'op-id': op_id,
-                       'object': u'{0}/{1}'.format(bucket, obj),
+                       'object': u'{0}/{1}'.format(bucket, obj.name),
                        'client-id': client_id,
                       }
                    )
@@ -194,54 +191,169 @@ def remove_op_state(connection, client_id, op_id, bucket, obj):
     return request(connection, 'delete', 'admin/opstate',
                    params={
                        'op-id': op_id,
-                       'object': u'{0}/{1}'.format(bucket, obj),
+                       'object': u'{0}/{1}'.format(bucket, obj.name),
                        'client-id': client_id,
                       },
                    expect_json=False,
                    )
 
+
 def get_bucket_list(connection):
     return list_metadata_keys(connection, 'bucket')
 
 
 @boto_call
-def list_objects_in_bucket(connection, bucket_name):
+def list_objects_in_bucket(connection, bucket_name, shard_id=-1, marker=None):
+    versioned = config['use_versioning']
     # use the boto library to do this
     bucket = connection.get_bucket(bucket_name)
+    list_call = bucket.list_versions if versioned else bucket.list
+    headers = None
+    if shard_id >= 0:
+        headers = {'Rgwx-Shard-Id': shard_id}
     try:
-        for key in bucket.list():
-            yield key.name
+        for key in list_call(headers=headers, marker=marker):
+            yield key
     except boto.exception.S3ResponseError as e:
         # since this is a generator, the exception will be raised when
         # it's read, rather than when this call returns, so raise a
         # unique exception to distinguish this from client errors from
         # other calls
         if e.status == 404:
-            raise BucketEmpty()
+            raise exc.BucketEmpty()
         else:
             raise
 
-
 @boto_call
-def delete_object(connection, bucket_name, object_name):
-    bucket = connection.get_bucket(bucket_name)
-    bucket.delete_key(object_name)
+def mark_delete_object(connection, bucket_name, obj, params=None):
+    """
+    Marking an object for deletion is only necessary for versioned objects, we
+    should not try these calls for non-versioned ones.
 
+    Usually, only full-sync operations will use this call, incremental should
+    perform actual delete operations with ``delete_versioned_object``
+    """
+    params = params or {}
+
+    params['rgwx-version-id'] = obj.version_id
+    params['rgwx-versioned-epoch'] = obj.VersionedEpoch
 
-def sync_object_intra_region(connection, bucket_name, object_name, src_zone,
-                             client_id, op_id):
     path = u'{bucket}/{object}'.format(
         bucket=bucket_name,
-        object=object_name,
+        object=obj.name,
         )
+
+    return request(connection, 'delete', path,
+                   params=params,
+                   expect_json=False)
+
+@boto_call
+def delete_versioned_object(connection, bucket_name, obj, mtime):
+    """
+    Perform a delete on a versioned object, the requirements for these types
+    of requests is to be able to pass the ``versionID`` as a query argument
+    """
+    # if obj.delete_marker is False we should not delete this and we shouldn't
+    # have been called, so return without doing anything
+    # if getattr(obj, 'delete_marker', False) is False:
+    #     log.info('obj: %s has `delete_marker=False`, will skip' % obj.name)
+    #     return
+
+    params = {}
+
+    if getattr(obj, 'version_id', None) is not None:
+        params['rgwx-version-id'] = obj.version_id
+        params['versionID'] = obj.version_id
+
+    if getattr(obj, 'VersionedEpoch', None) is not None:
+        params['rgwx-versioned-epoch'] = obj.VersionedEpoch
+
+    headers = {}
+    if mtime is not None:
+        params['no-precondition-error'] = 'true'
+        headers['X-Amz-Delete-If-Unmodified-Since'] = mtime
+
+    path = u'{bucket}/{object}'.format(
+        bucket=bucket_name,
+        object=obj.name,
+        )
+
+    return request(connection, 'delete', path,
+                   params=params,
+                   headers=headers,
+                   expect_json=False)
+
+
+@boto_call
+def delete_object(connection, bucket_name, obj, mtime=None):
+    delete_versioned_object(connection, bucket_name, obj, mtime)
+
+def is_versioned(obj):
+    """
+    Check if a given object is versioned by inspecting some of its attributes.
+    """
+    # before any heuristic, newer versions of RGW will tell if an obj is
+    # versioned so try that first
+    if hasattr(obj, 'versioned'):
+        return obj.versioned
+
+    if hasattr(obj, 'VersionedEpoch'):
+        if getattr(obj, 'VersionedEpoch', 0) == 0:
+            return False
+        # overly paranoid here, an object that is not versioned should *never*
+        # have a `VersionedEpoch` attribute
+        return True
+    else:
+        if getattr(obj, 'version_id', None):
+            if obj.version_id is None:
+                return False
+            return True  # probably will never get here
+
+    return False
+
+
+def sync_object_intra_region(connection, bucket_name, bucket_id, obj, src_zone,
+                             client_id, op_id):
+
+    params = {
+        'rgwx-source-zone': src_zone,
+        'rgwx-client-id': client_id,
+        'rgwx-op-id': op_id,
+    }
+
+    if bucket_id is not None:
+        params['rgwx-bucket-instance'] = bucket_id
+
+    if is_versioned(obj):
+        log.debug('detected obj as versioned: %s' % obj.name)
+        log.debug('obj attributes are:')
+        for k in dir(obj):
+            if not k.startswith('_'):
+                v = getattr(obj, k, None)
+                log.debug('%s.%s = %s' % (obj.name, k, v))
+
+        # set the extra params to support versioned operations
+        params['rgwx-version-id'] = obj.version_id
+        params['rgwx-versioned-epoch'] = obj.VersionedEpoch
+
+        # delete_marker may not exist in the obj
+        if getattr(obj, 'delete_marker', None) is True:
+            log.debug('obj %s has a delete_marker, marking for deletion' % obj.name)
+            # when the object has a delete marker we need to create it with
+            # a delete marker on the destination rather than copying
+            return mark_delete_object(connection, bucket_name, obj, params=params)
+
+    path = u'{bucket}/{object}'.format(
+        bucket=bucket_name,
+        object=obj.name,
+        )
+
+    params['rgwx-copy-if-newer'] = 'true'
+
     return request(connection, 'put', path,
-                   params={
-                       'rgwx-source-zone': src_zone,
-                       'rgwx-client-id': client_id,
-                       'rgwx-op-id': op_id,
-                       },
+                   params=params,
                    headers={
-                       'x-amz-copy-source': url_safe('%s/%s' % (bucket_name, object_name)),
+                       'x-amz-copy-source': url_safe('%s/%s' % (bucket_name, obj.name)),
                        },
                    expect_json=False)
 
@@ -305,55 +417,109 @@ def num_log_shards(connection, shard_type):
 
 
 def set_worker_bound(connection, type_, marker, timestamp,
-                     daemon_id, id_, data=None):
+                     daemon_id, id_, data=None, sync_type='incremental', key=None):
+    """
+
+    :param sync_type: The type of synchronization that should be attempted by
+    the agent, defaulting to "incremental" but can also be "full".
+    """
     if data is None:
         data = []
-    key = _id_name(type_)
     boto.log.debug('set_worker_bound: data = %r', data)
-    return request(
-        connection, 'post', 'admin/replica_log',
-        params={
+    params={
             'type': type_,
-            key: id_,
             'marker': marker,
             'time': timestamp,
             'daemon_id': daemon_id,
-            },
+            'sync-type': sync_type,
+            }
+    if id_ is not None:
+        params[_id_name(type_)] = id_
+    if key is not None:
+        params['key'] = key
+    return request(
+        connection, 'post', 'admin/replica_log',
+        params=params,
         data=json.dumps(data),
         special_first_param='work_bound',
         )
 
 
-def del_worker_bound(connection, type_, daemon_id, id_):
-    key = _id_name(type_)
+def del_worker_bound(connection, type_, daemon_id, id_, key = None):
+    params={
+        'type': type_,
+        'daemon_id': daemon_id,
+        }
+    if id_ is not None:
+        params[_id_name(type_)] = id_
+    if key is not None:
+        params['key'] = key
     return request(
         connection, 'delete', 'admin/replica_log',
-        params={
-            'type': type_,
-            key: id_,
-            'daemon_id': daemon_id,
-            },
+        params=params,
         special_first_param='work_bound',
         expect_json=False,
         )
 
 
-def get_worker_bound(connection, type_, id_):
-    key = _id_name(type_)
-    out = request(
-        connection, 'get', 'admin/replica_log',
-        params={
-            'type': type_,
-            key: id_,
-            },
-        special_first_param='bounds',
-        )
-    boto.log.debug('get_worker_bound returned: %r', out)
+def get_worker_bound(connection, type_, id_, init_if_not_found=True, key = None):
+    params={
+        'type': type_,
+        }
+    if id_ is not None:
+        params[_id_name(type_)] = id_
+    if key is not None:
+        params['key'] = key
+    try:
+        out = request(
+            connection, 'get', 'admin/replica_log',
+            params=params,
+            special_first_param='bounds',
+            )
+        dev_log.debug('get_worker_bound returned: %r', out)
+    except exc.NotFound:
+        dev_log.debug('no worker bound found for bucket instance "%s"',
+                      id_)
+        # if no worker bounds have been set, start from the beginning
+        # returning fallback, default values
+        if init_if_not_found:
+            return dict(
+                marker=' ',
+                oldest_time=DEFAULT_TIME,
+                retries=[]
+            )
+
+        return None
+
     retries = set()
     for item in out['markers']:
         names = [retry['name'] for retry in item['items_in_progress']]
         retries = retries.union(names)
-    return out['marker'], out['oldest_time'], retries
+    out['retries'] = retries
+    return out
+
+def list_worker_bound(connection, type_, id_):
+    params={
+        'type': type_,
+        }
+    if id_ is not None:
+        params[_id_name(type_)] = id_
+    params['list-keys'] = True
+    try:
+        out = request(
+            connection, 'get', 'admin/replica_log',
+            params=params,
+            special_first_param='bounds',
+            )
+        dev_log.debug('get_worker_bound returned: %r', out)
+    except exc.NotFound:
+        dev_log.debug('no worker bound found for %s "%s"',
+                      type_, id_)
+        # if no worker bounds have been set, start from the beginning
+        # returning fallback, default values
+        return None
+
+    return out
 
 
 class Zone(object):
@@ -414,7 +580,7 @@ class RegionMap(object):
             for zone in region.zones.itervalues():
                 if endpoint in zone.endpoints or endpoint.zone == zone.name:
                     return region, zone
-        raise ZoneNotFound('%s not found in region map' % endpoint)
+        raise exc.ZoneNotFound('%s not found in region map' % endpoint)
 
 
 def get_region_map(connection):
@@ -424,24 +590,25 @@ def get_region_map(connection):
 
 def _validate_sync_dest(dest_region, dest_zone):
     if dest_region.is_master and dest_zone.is_master:
-        raise InvalidZone('destination cannot be master zone of master region')
+        raise exc.InvalidZone('destination cannot be master zone of master region')
 
 
 def _validate_sync_source(src_region, src_zone, dest_region, dest_zone,
                           meta_only):
     if not src_zone.is_master:
-        raise InvalidZone('source zone %s must be a master zone' % src_zone.name)
+        raise exc.InvalidZone('source zone %s must be a master zone' % src_zone.name)
     if (src_region.name == dest_region.name and
         src_zone.name == dest_zone.name):
-        raise InvalidZone('source and destination must be different zones')
+        raise exc.InvalidZone('source and destination must be different zones')
     if not src_zone.log_meta:
-        raise InvalidZone('source zone %s must have metadata logging enabled' % src_zone.name)
+        raise exc.InvalidZone('source zone %s must have metadata logging enabled' % src_zone.name)
     if not meta_only and not src_zone.log_data:
-        raise InvalidZone('source zone %s must have data logging enabled' % src_zone.name)
+        raise exc.InvalidZone('source zone %s must have data logging enabled' % src_zone.name)
     if not meta_only and src_region.name != dest_region.name:
-        raise InvalidZone('data sync can only occur between zones in the same region')
+        raise exc.InvalidZone('data sync can only occur between zones in the same region')
     if not src_zone.endpoints:
-        raise InvalidZone('region map contains no endpoints for default source zone %s' % src_zone.name)
+        raise exc.InvalidZone('region map contains no endpoints for default source zone %s' % src_zone.name)
+
 
 def configure_endpoints(region_map, dest_endpoint, src_endpoint, meta_only):
     print('region map is: %r' % region_map)
@@ -459,7 +626,7 @@ def configure_endpoints(region_map, dest_endpoint, src_endpoint, meta_only):
             _validate_sync_source(dest_region, dest_region.master_zone,
                                   dest_region, dest_zone, meta_only)
             src_region, src_zone = dest_region, dest_region.master_zone
-        except InvalidZone as e:
+        except exc.InvalidZone as e:
             log.debug('source region %s zone %s unaccetpable: %s',
                       dest_region.name, dest_region.master_zone.name, e)
             master_region = region_map.master_region
@@ -481,6 +648,7 @@ def configure_endpoints(region_map, dest_endpoint, src_endpoint, meta_only):
     src_endpoint.region = src_region
     src_endpoint.zone = src_zone
 
+
 class S3ConnectionWrapper(object):
     def __init__(self, endpoint, debug):
         self.endpoint = endpoint
@@ -509,5 +677,7 @@ class S3ConnectionWrapper(object):
     def __getattr__(self, attrib):
         return getattr(self.s3_connection, attrib)
 
+
 def connection(endpoint, debug=None):
+    log.info('creating connection to endpoint: %s' % endpoint)
     return S3ConnectionWrapper(endpoint, debug)
