@@ -5,6 +5,8 @@ import httpretty
 import re
 
 from radosgw_agent import client
+from radosgw_agent import exceptions as exc
+from radosgw_agent.constants import DEFAULT_TIME
 
 # parametrization helpers
 
@@ -209,9 +211,9 @@ def test_parse_repr(url, host, port, secure):
 
 
 def test_parse_endpoint_bad_input():
-    with py.test.raises(client.InvalidProtocol):
+    with py.test.raises(exc.InvalidProtocol):
         client.parse_endpoint('ftp://example.com')
-    with py.test.raises(client.InvalidHost):
+    with py.test.raises(exc.InvalidHost):
         client.parse_endpoint('http://:80/')
 
 def _test_configure_endpoints(dest_url, dest_region, dest_zone,
@@ -237,7 +239,7 @@ def test_configure_endpoints_2nd_region_master_zone_meta():
                               meta_only=True)
 
 def test_configure_endpoints_2nd_region_master_zone_data():
-    with py.test.raises(client.InvalidZone):
+    with py.test.raises(exc.InvalidZone):
         _test_configure_endpoints('http://vit:8000', 'swab', 'swab-1',
                                   'http://vit:8001', 'skinny', 'skinny-1',
                                   meta_only=False)
@@ -256,7 +258,7 @@ def test_configure_endpoints_2nd_region_readonly_meta():
                               meta_only=True)
 
 def test_configure_endpoints_2nd_region_readonly_data():
-    with py.test.raises(client.InvalidZone):
+    with py.test.raises(exc.InvalidZone):
         _test_configure_endpoints('http://ro:8080', 'readonly', 'ro-2',
                                   'http://vit:8001', 'skinny', 'skinny-1',
                                   meta_only=False)
@@ -267,13 +269,13 @@ def test_configure_endpoints_2nd_region_metaonly_meta():
                               meta_only=True)
 
 def test_configure_endpoints_2nd_region_metaonly_data():
-    with py.test.raises(client.InvalidZone):
+    with py.test.raises(exc.InvalidZone):
         _test_configure_endpoints('http://meta:8080', 'metaonly', 'meta-2',
                                   'http://vit:8001', 'skinny', 'skinny-1',
                                   meta_only=False)
 
 def test_configure_endpoints_master_region_master_zone():
-    with py.test.raises(client.InvalidZone):
+    with py.test.raises(exc.InvalidZone):
         _test_configure_endpoints('http://vit:8001', 'skinny', 'skinny-1',
                                   'http://vit:8001', 'skinny', 'skinny-1')
 
@@ -288,37 +290,37 @@ def test_configure_endpoints_specified_src_master_region_meta():
                               'http://vit:8001', meta_only=True)
 
 def test_configure_endpoints_specified_src_master_region_data():
-    with py.test.raises(client.InvalidZone):
+    with py.test.raises(exc.InvalidZone):
         _test_configure_endpoints('http://vit:8003', 'swab', 'swab-2',
                                   'http://vit:8001', 'skinny', 'skinny-1',
                                   'http://vit:8001', meta_only=False)
 
 def test_configure_endpoints_bad_src_same_region():
-    with py.test.raises(client.InvalidZone):
+    with py.test.raises(exc.InvalidZone):
         _test_configure_endpoints('http://vit:8003', 'swab', 'swab-2',
                                   'http://vit:8004', 'swab', 'swab-3',
                                   'http://vit:8004')
 
 def test_configure_endpoints_bad_src_master_region():
-    with py.test.raises(client.InvalidZone):
+    with py.test.raises(exc.InvalidZone):
         _test_configure_endpoints('http://vit:8003', 'swab', 'swab-2',
                                   'http://vit:8002', 'skinny', 'skinny-2',
                                   'http://vit:8002')
 
 def test_configure_endpoints_bad_src_same_zone():
-    with py.test.raises(client.InvalidZone):
+    with py.test.raises(exc.InvalidZone):
         _test_configure_endpoints('http://vit:8000', 'swab', 'swab-1',
                                   'http://vit:8000', 'swab', 'swab-1',
                                   'http://vit:8000')
 
 def test_configure_endpoints_specified_nonexistent_src():
-    with py.test.raises(client.ZoneNotFound):
+    with py.test.raises(exc.ZoneNotFound):
         _test_configure_endpoints('http://vit:8005', 'skinny', 'skinny-1',
                                   'http://vit:8001', 'skinny', 'skinny-1',
                                   'http://vit:80')
 
 def test_configure_endpoints_unknown_zone():
-    with py.test.raises(client.ZoneNotFound):
+    with py.test.raises(exc.ZoneNotFound):
         _test_configure_endpoints('http://vit:8005', 'skinny', 'skinny-1',
                                   'http://vit:8001', 'skinny', 'skinny-1')
 
@@ -338,7 +340,7 @@ class TestCheckResultStatus(object):
     def test_check_raises_http_error(self, code):
         response = Mock()
         response.status = code
-        with py.test.raises(client.HttpError):
+        with py.test.raises(exc.HttpError):
             client.check_result_status(response)
 
     @py.test.mark.parametrize('code', http_valid_status_codes())
@@ -351,7 +353,7 @@ class TestCheckResultStatus(object):
     def test_check_raises_not_found(self):
         response = Mock()
         response.status = 404
-        with py.test.raises(client.NotFound):
+        with py.test.raises(exc.NotFound):
             client.check_result_status(response)
 
 
@@ -369,7 +371,7 @@ class TestBotoCall(object):
         def foo():
             raise boto.exception.S3ResponseError(404, '')
 
-        with py.test.raises(client.NotFound):
+        with py.test.raises(exc.NotFound):
             foo()
 
     def test_non_boto_exception(self):
@@ -433,7 +435,7 @@ class TestRequest(object):
             True,
         )
 
-        with py.test.raises(client.HttpError):
+        with py.test.raises(exc.HttpError):
             client.request(connection, 'get', '/%7E~', _retries=0)
 
 
@@ -451,7 +453,7 @@ class TestBotoCall(object):
         def foo():
             raise boto.exception.S3ResponseError(404, '')
 
-        with py.test.raises(client.NotFound):
+        with py.test.raises(exc.NotFound):
             foo()
 
     def test_non_boto_exception(self):
@@ -463,7 +465,7 @@ class TestBotoCall(object):
             foo()
 
 
-class TestGETClientRequests(object):
+class TestGETClientRequestsPaths(object):
 
     def setup(self):
         self.connection = client.connection(
@@ -471,11 +473,12 @@ class TestGETClientRequests(object):
             True,
         )
 
-    def register(self):
+    def register(self, body=None):
+        body = body or '{}'
         httpretty.register_uri(
             httpretty.GET,
             re.compile("http://localhost:8888/(.*)"),
-            body='{}',
+            body=body,
             content_type="application/json",
         )
 
@@ -489,7 +492,6 @@ class TestGETClientRequests(object):
     @httpretty.activate
     def test_get_metadata_no_re_encoding(self):
         self.register()
-        #client.get_metadata(self.connection, 'bucket.instance', 'mybar%3Ar0z0.4140.1')
         client.get_metadata(self.connection, 'bucket.instance', 'mybar:r0z0.4140.1')
         server_request = httpretty.last_request()
         assert server_request.path == '/admin/metadata/bucket.instance?key=mybar%3Ar0z0.4140.1'
@@ -515,7 +517,6 @@ class TestGETClientRequests(object):
         server_request = httpretty.last_request()
         assert server_request.path == '/admin/metadata/bucket'
 
-
     @httpretty.activate
     def test_url_response(self):
 
@@ -527,3 +528,145 @@ class TestGETClientRequests(object):
         )
         result = client.request(self.connection, 'get', '/%7E~')
         assert result == {'msg': 'ok'}
+
+
+class TestClientListObjectsInBucket(object):
+
+    def setup(self):
+        self.connection = client.connection(
+            client.Endpoint('localhost', 8888, False, 'key', 'secret'),
+            True,
+        )
+        self.body = """
+        [
+            {
+                "name": "mahobject/",
+                "etag": "d41d8cd98f00b204e9800998ecf8427e",
+                "content_type": "application/octet-stream",
+                "last_modified": "2015-01-15T15:24:42.000Z",
+                "storage_class": "STANDARD",
+                "owner": {
+                    "display_name": "client1-system-user",
+                    "id": "client1-system-user"
+                }
+            }
+        ]
+        """
+
+    def register(self, body=None):
+        body = body or self.body
+        httpretty.register_uri(
+            httpretty.GET,
+            re.compile("http://localhost:8888/(.*)"),
+            body=body,
+            content_type="application/json",
+        )
+
+    @httpretty.activate
+    def test_get_bucket_is_a_single_item(self):
+        self.register()
+        result = client.get_bucket_list(self.connection)
+        assert len(result) == 1
+
+    @httpretty.activate
+    def test_get_bucket_has_right_metadata(self):
+        self.register()
+        result = client.get_bucket_list(self.connection)
+        obj = result[0]
+        owner = {
+            "display_name": "client1-system-user",
+            "id": "client1-system-user"
+        }
+        assert obj['name'] == 'mahobject/'
+        assert obj['etag'] == 'd41d8cd98f00b204e9800998ecf8427e'
+        assert obj['content_type'] == 'application/octet-stream'
+        assert obj['last_modified'] == '2015-01-15T15:24:42.000Z'
+        assert obj['storage_class'] == 'STANDARD'
+        assert obj['owner'] == owner
+
+
+class TestClientGetWorkerBound(object):
+
+    def setup(self):
+        self.connection = client.connection(
+            client.Endpoint('localhost', 8888, False, 'key', 'secret'),
+            True,
+        )
+        self.body = """
+        {"marker": "00000000002.2.3",
+            "markers": [
+                {
+                    "entity": "radosgw-agent",
+                    "items_in_progress": [
+                        {
+                            "name": "hello",
+                            "timestamp": "0.000000"
+                        }
+                    ],
+                    "position_marker": "00000000002.2.3",
+                    "position_time": "0.000000"
+                }
+            ],
+         "oldest_time": "0.000000"
+        }
+        """
+
+    def register(self, body=None, status=200):
+        body = body or self.body
+        httpretty.register_uri(
+            httpretty.GET,
+            re.compile("http://localhost:8888/(.*)"),
+            body=body,
+            content_type="application/json",
+            status=status
+        )
+
+    @httpretty.activate
+    def test_get_bound_has_right_metadata(self):
+        self.register()
+        result = client.get_worker_bound(
+            self.connection,
+            'bucket-index',
+            'beast:us-east'
+        )
+        assert result['marker'] == "00000000002.2.3"
+        assert result['retries'] == set(['hello'])
+        assert result['oldest_time'] == "0.000000"
+
+    @httpretty.activate
+    def test_get_bound_fails_fallsback_to_defaults(self):
+        self.register(status=404)
+        result = client.get_worker_bound(
+            self.connection,
+            'bucket-index',
+            'beast:us-east'
+        )
+        assert result['marker'] == " "
+        assert result['retries'] == []
+        assert result['oldest_time'] == DEFAULT_TIME
+
+
+class TestIsVersioned(object):
+
+    def setup(self):
+        # set strict attributes in the mock
+        self.obj = Mock(spec=object)
+
+    def test_is_in_fact_versioned(self):
+        self.obj.VersionedEpoch = u'1'
+        self.obj.version_id = 'somehashvalue'
+        assert client.is_versioned(self.obj) is True
+
+    def test_is_not_versioned_no_attr_versioned_epoch(self):
+        assert client.is_versioned(self.obj) is False
+
+    def test_is_not_versioned_no_attr_version_id(self):
+        assert client.is_versioned(self.obj) is False
+
+    def test_is_versioned_version_id(self):
+        self.obj.version_id = 1
+        assert client.is_versioned(self.obj) is True
+
+    def test_is_not_versioned_versioned_id_is_none(self):
+        self.obj.version_id = None
+        assert client.is_versioned(self.obj) is False
