@@ -1,4 +1,5 @@
 from collections import namedtuple
+from itertools import ifilter
 import logging
 import multiprocessing
 import os
@@ -7,23 +8,13 @@ import time
 
 from radosgw_agent import client
 from radosgw_agent import lock
+from radosgw_agent.util import obj as obj_, get_dev_logger
+from radosgw_agent.exceptions import SkipShard, SyncError, SyncTimedOut, SyncFailed, NotFound, BucketEmpty
+from radosgw_agent.constants import DEFAULT_TIME, RESULT_SUCCESS, RESULT_ERROR
 
 log = logging.getLogger(__name__)
+dev_log = get_dev_logger(__name__)
 
-RESULT_SUCCESS = 0
-RESULT_ERROR = 1
-
-class SkipShard(Exception):
-    pass
-
-class SyncError(Exception):
-    pass
-class SyncTimedOut(SyncError):
-    pass
-class SyncFailed(SyncError):
-    pass
-
-DEFAULT_TIME = '1970-01-01 00:00:00'
 
 class Worker(multiprocessing.Process):
     """sync worker to run in its own process"""
@@ -56,7 +47,7 @@ class Worker(multiprocessing.Process):
         try:
             self.lock.set_shard(shard_num)
             self.lock.acquire()
-        except client.NotFound:
+        except NotFound:
             # no log means nothing changed this shard yet
             self.lock.unset_shard()
             self.result_queue.put((RESULT_SUCCESS, result))
@@ -78,17 +69,19 @@ class Worker(multiprocessing.Process):
             log.warn('error unlocking log, continuing anyway '
                      'since lock will timeout. Traceback:', exc_info=True)
 
-    def set_bound(self, key, marker, retries, type_=None):
+    def set_bound(self, key, marker, retries, timestamp, type_=None):
         # api doesn't allow setting a bound with a blank marker
         if marker:
             if type_ is None:
                 type_ = self.type
             try:
-                data = [dict(name=item, time=DEFAULT_TIME) for item in retries]
+                data = [
+                    obj_.to_dict(item, time=DEFAULT_TIME) for item in retries
+                ]
                 client.set_worker_bound(self.dest_conn,
                                         type_,
                                         marker,
-                                        DEFAULT_TIME,
+                                        timestamp,
                                         self.daemon_id,
                                         key,
                                         data=data)
@@ -102,6 +95,7 @@ class Worker(multiprocessing.Process):
 MetadataEntry = namedtuple('MetadataEntry',
                            ['section', 'name', 'marker', 'timestamp'])
 
+
 def _meta_entry_from_json(entry):
     return MetadataEntry(
         entry['section'],
@@ -111,15 +105,69 @@ def _meta_entry_from_json(entry):
         )
 
 BucketIndexEntry = namedtuple('BucketIndexEntry',
-                              ['object', 'marker', 'timestamp'])
+                              [
+                                  'object',
+                                  'marker',
+                                  'timestamp',
+                                  'op',
+                                  'versioned',
+                                  'ver',
+                                  'name',
+                                  # compatibility with boto objects:
+                                  'VersionedEpoch',
+                                  'version_id',
+                              ])
+
+BucketVer = namedtuple('BucketVer',
+        [
+            'epoch',
+            'pool',
+        ])
+
 
 def _bi_entry_from_json(entry):
+    ver = entry.get('ver', {})
+    entry_ver = BucketVer(
+        ver.get('epoch'),
+        ver.get('pool')
+    )
+
+    # compatibility with boto objects:
+    VersionedEpoch = ver.get('epoch')
+    version_id = entry.get('instance', 'null')
+
     return BucketIndexEntry(
         entry['object'],
         entry['op_id'],
         entry['timestamp'],
+        entry.get('op', ''),
+        entry.get('versioned', False),
+        entry_ver,
+        entry['object'],
+        VersionedEpoch,
+        version_id,
         )
 
+
+def filter_versioned_objects(entry):
+    """
+    On incremental sync operations, the log may indicate that 'olh' entries,
+    which should be ignored. So this filter function will check for the
+    different attributes present in an ``entry`` and return only valid ones.
+
+    This should be backwards compatible with older gateways that return log
+    entries that don't support versioning.
+    """
+    # do not attempt filtering on non-versioned entries
+    if not entry.versioned:
+        return entry
+
+    # writes or delete 'op' values should be ignored
+    if entry.op not in ['write', 'delete']:
+        # allowed op states are `link_olh` and `link_olh_del`
+        return entry
+
+
 class IncrementalMixin(object):
     """This defines run() and get_and_process_entries() for incremental sync.
 
@@ -132,7 +180,7 @@ class IncrementalMixin(object):
         while True:
             item = self.work_queue.get()
             if item is None:
-                log.info('process %s is done. Exiting', self.ident)
+                dev_log.info('process %s is done. Exiting', self.ident)
                 break
 
             shard_num, (log_entries, retries) = item
@@ -170,35 +218,41 @@ class DataWorker(Worker):
         self.object_sync_timeout = kwargs.get('object_sync_timeout', 60 * 60 * 60)
         self.daemon_id = kwargs['daemon_id']
 
-    def sync_object(self, bucket, obj):
-        log.debug('sync_object %s/%s', bucket, obj)
+    def sync_object(self, bucket, bucket_id, obj):
+        log.debug('syncing object %s/%s', bucket, obj.name)
         self.op_id += 1
         local_op_id = self.local_lock_id + ':' +  str(self.op_id)
         found = False
 
         try:
             until = time.time() + self.object_sync_timeout
-            client.sync_object_intra_region(self.dest_conn, bucket, obj,
+            client.sync_object_intra_region(self.dest_conn, bucket,
+                                            bucket_id,
+                                            obj,
                                             self.src.zone.name,
                                             self.daemon_id,
                                             local_op_id)
             found = True
-        except client.NotFound:
-            log.debug('"%s/%s" not found on master, deleting from secondary',
-                      bucket, obj)
+        except NotFound:
+            log.debug('object "%s/%s" not found on master, deleting from secondary',
+                      bucket, obj.name)
             try:
                 client.delete_object(self.dest_conn, bucket, obj)
-            except client.NotFound:
+                client.remove_op_state(self.dest_conn, self.daemon_id,
+                                       local_op_id, bucket, obj)
+            except NotFound:
                 # Since we were trying to delete the object, just return
                 return False
             except Exception:
-                msg = 'could not delete "%s/%s" from secondary' % (bucket, obj)
+                msg = 'could not delete "%s/%s" from secondary' % (bucket, obj.name)
                 log.exception(msg)
                 raise SyncFailed(msg)
         except SyncFailed:
             raise
-        except Exception as e:
-            log.warn('encountered an exception during sync', exc_info=True)
+        except Exception as error:
+            msg = 'encountered an error during sync'
+            dev_log.warn(msg, exc_info=True)
+            log.warning('%s: %s' % (msg, error))
             # wait for it if the op state is in-progress
             self.wait_for_object(bucket, obj, until, local_op_id)
         # TODO: clean up old op states
@@ -206,7 +260,7 @@ class DataWorker(Worker):
             if found:
                 client.remove_op_state(self.dest_conn, self.daemon_id,
                                        local_op_id, bucket, obj)
-        except client.NotFound:
+        except NotFound:
             log.debug('op state already gone')
         except Exception:
             log.exception('could not remove op state for daemon "%s" op_id %s',
@@ -214,6 +268,22 @@ class DataWorker(Worker):
 
         return True
 
+    def delete_object(self, bucket, obj, mtime):
+        log.debug('syncing object %s/%s', bucket, obj.name)
+        self.op_id += 1
+        local_op_id = self.local_lock_id + ':' +  str(self.op_id)
+
+        try:
+            client.delete_object(self.dest_conn, bucket, obj, mtime)
+        except NotFound:
+            log.debug('object "%s/%s" not found', bucket, obj.name)
+        except Exception as error:
+            msg = 'encountered an error during sync'
+            dev_log.warn(msg, exc_info=True)
+            log.warning('%s: %s' % (msg, error))
+
+        return True
+
     def wait_for_object(self, bucket, obj, until, local_op_id):
         while time.time() < until:
             try:
@@ -230,7 +300,7 @@ class DataWorker(Worker):
                 time.sleep(1)
             except SyncFailed:
                 raise
-            except client.NotFound:
+            except NotFound:
                 raise SyncFailed('object copy state not found')
             except Exception as e:
                 log.debug('error geting op state: %s', e, exc_info=True)
@@ -242,29 +312,33 @@ class DataWorker(Worker):
         metadata = client.get_metadata(self.src_conn, 'bucket', bucket)
         return bucket + ':' + metadata['data']['bucket']['bucket_id']
 
+    def get_bucket_metadata(self, bucket):
+        bucket_instance = self.get_bucket_instance(bucket)
+        metadata = client.get_metadata(self.src_conn, 'bucket.instance', bucket_instance)
+        return bucket_instance, metadata
+
     def get_bucket(self, bucket_instance):
         return bucket_instance.split(':', 1)[0]
 
     def sync_bucket(self, bucket, objects):
+        log.info('*'*80)
         log.info('syncing bucket "%s"', bucket)
         retry_objs = []
         count = 0
         for obj in objects:
-            count += 1
-            # sync each object
-            log.debug('syncing object "%s/%s"', bucket, obj),
             try:
                 self.sync_object(bucket, obj)
+                count += 1
             except SyncError as err:
                 log.error('failed to sync object %s/%s: %s',
-                          bucket, obj, err)
+                          bucket, obj.name, err)
+                log.warning(
+                    'will retry sync of failed object at next incremental sync'
+                )
                 retry_objs.append(obj)
-
-        log.debug('bucket {bucket} has {num_objects} object'.format(
-                  bucket=bucket, num_objects=count))
-        if retry_objs:
-            log.debug('these objects failed to be synced and will be during '
-                      'the next incremental sync: %s', retry_objs)
+        log.info('synced %s objects' % count)
+        log.info('completed syncing bucket "%s"', bucket)
+        log.info('*'*80)
 
         return retry_objs
 
@@ -281,7 +355,7 @@ class DataWorkerIncremental(IncrementalMixin, DataWorker):
             try:
                 log_entries = client.get_log(self.src_conn, 'bucket-index',
                                              marker, self.max_entries, instance)
-            except client.NotFound:
+            except NotFound:
                 log_entries = []
 
             log.debug('bucket instance "%s" has %d entries after "%s"', instance,
@@ -304,11 +378,15 @@ class DataWorkerIncremental(IncrementalMixin, DataWorker):
 
     def inc_sync_bucket_instance(self, instance, marker, timestamp, retries):
         max_marker, entries = self.get_bucket_instance_entries(marker, instance)
-        objects = set([entry.object for entry in entries])
+
+        # regardless if entries are versioned, make sure we filter them
+        entries = [i for i in ifilter(filter_versioned_objects, entries)]
+
+        objects = set([entry for entry in entries])
         bucket = self.get_bucket(instance)
         new_retries = self.sync_bucket(bucket, objects.union(retries))
 
-        result = self.set_bound(instance, max_marker, new_retries,
+        result = self.set_bound(instance, max_marker, new_retries, timestamp,
                                 'bucket-index')
         if new_retries:
             result = RESULT_ERROR
@@ -326,15 +404,17 @@ class DataWorkerIncremental(IncrementalMixin, DataWorker):
             if ':' not in bucket_instance:
                 # it's just a plain bucket from an old version of the agent
                 bucket_instance = self.get_bucket_instance(bucket_instance)
-            try:
-                marker, timestamp, retries = client.get_worker_bound(
-                    self.dest_conn,
-                    'bucket-index',
-                    bucket_instance)
-            except client.NotFound:
-                log.debug('no worker bound found for bucket instance "%s"',
-                          bucket_instance)
-                marker, timestamp, retries = ' ', DEFAULT_TIME, []
+
+            bound = client.get_worker_bound(
+                self.dest_conn,
+                'bucket-index',
+                bucket_instance)
+
+            marker = bound['marker']
+            # remap dictionaries to object-like
+            retries = [obj_.to_obj(i) for i in bound['retries']]
+            timestamp = bound['oldest_time']
+
             try:
                 sync_result = self.inc_sync_bucket_instance(bucket_instance,
                                                             marker,
@@ -349,6 +429,7 @@ class DataWorkerIncremental(IncrementalMixin, DataWorker):
 
         return new_retries
 
+
 class DataWorkerFull(DataWorker):
 
     def full_sync_bucket(self, bucket):
@@ -357,16 +438,16 @@ class DataWorkerFull(DataWorker):
             try:
                 marker = client.get_log_info(self.src_conn, 'bucket-index',
                                              instance)['max_marker']
-            except client.NotFound:
+            except NotFound:
                 marker = ' '
             log.debug('bucket instance is "%s" with marker %s', instance, marker)
 
             objects = client.list_objects_in_bucket(self.src_conn, bucket)
             retries = self.sync_bucket(bucket, objects)
 
-            result = self.set_bound(instance, marker, retries, 'bucket-index')
+            result = self.set_bound(instance, marker, retries, DEFAULT_TIME, 'bucket-index')
             return not retries and result == RESULT_SUCCESS
-        except client.BucketEmpty:
+        except BucketEmpty:
             log.debug('no objects in bucket %s', bucket)
             return True
         except Exception:
@@ -401,8 +482,10 @@ class DataWorkerFull(DataWorker):
             self.unlock_shard()
             self.result_queue.put((RESULT_SUCCESS, (shard_num, retry_buckets)))
             log.info('finished syncing shard %d', shard_num)
-            log.info('incremental sync will need to retry buckets: %s',
-                     retry_buckets)
+            if retry_buckets:
+                log.info('incremental sync will need to retry buckets: %s',
+                         retry_buckets)
+
 
 class MetadataWorker(Worker):
 
@@ -414,12 +497,12 @@ class MetadataWorker(Worker):
         log.debug('syncing metadata type %s key "%s"', section, name)
         try:
             metadata = client.get_metadata(self.src_conn, section, name)
-        except client.NotFound:
+        except NotFound:
             log.debug('%s "%s" not found on master, deleting from secondary',
                       section, name)
             try:
                 client.delete_metadata(self.dest_conn, section, name)
-            except client.NotFound:
+            except NotFound:
                 # Since this error is handled appropriately, return success
                 return RESULT_SUCCESS
         except Exception as e:
